library collisionsxy;

import 'dart:math' as math;
import 'package:vector_math/vector_math.dart';

class CollisionInfo {
 final normal = new vec2.zero();
 var depth = 0.0;
 /// edge of collision start point
 vec3 edgev0 = null;
/// edge of collision end point
 vec3 edgev1 = null;
 /// collision vertex
 vec3 v;
}

class MinMax {
  double min;
  double max;
}

void projectToAxis( List<vec3> vertices, vec3 axis, MinMax out) {
  var dotP = axis.dot(vertices[0]);

  //Set the minimum and maximum values to the projection of the first vertex
  out.min = out.max = dotP;

  for( int i = vertices.length - 1 ; i > 1 ; --i) {
    //Project the rest of the vertices onto the axis and extend
    //the interval to the left/right if necessary
    dotP = axis.dot(vertices[i]);
    out.min = math.min( dotP, out.min );
    out.max = math.max( dotP, out.max );
  }
}

/// Calculate the distance between the two intervals
double intervalDistance( double minA, double maxA, double minB, double maxB ) {
  return ( minA < minB ) ? minB - maxA : minA - maxB;
}

//TODO remove the need of [m1] and [m2]
double intervalDistanceOnAxis(List<vec3> shapeA, List<vec3> shapeB, vec3 axis, MinMax m1, MinMax m2) {
  // Project both bodies onto the perpendicular
  projectToAxis(shapeA, axis, m1);
  projectToAxis(shapeB, axis, m2);

  return intervalDistance( m1.min, m1.max, m2.min, m2.max );
}

/// Used for optimization to test if the bounding boxes (aabb) of two bodies overlap.
/// bounding boxes are vec4(minX, minY, width, height)
bool detectCollision_aabb(vec4 shapeA, vec4 shapeB) {
  return ( shapeA.x <= (shapeB.x + shapeB.z)) && ( shapeA.y <= (shapeB.y + shapeB.w)) && ((shapeA.x + shapeA.z) >= shapeB.x ) && (( shapeB.y + shapeB.w) >= shapeA.y);
}
/// source port from
/// * [A Verlet based approach for 2D game physics](http://www.gamedev.net/page/resources/_/technical/math-and-physics/a-verlet-based-approach-for-2d-game-physics-r2714)
/// * [java port](https://code.google.com/p/verlet/source/browse/VerletDemo/src/com/gmail/craigmit/verletdemo/Physics.java)
bool detectCollision_sat( List<vec3> shapeA, List<vec3> shapeB, CollisionInfo out, vec3 axis ) {
  var m1 = new MinMax();
  var m2 = new MinMax();
  return _detectCollision_satEdge1(shapeA, shapeB, out, axis, m1, m2)
      || _detectCollision_satEdge1(shapeB, shapeA, out, axis, m1, m2)
      ;
}

/// The body containing the collision edge lies in [shapeB] and the one conatining the collision vertex in [shapeA]
bool _detectCollision_satEdge1( List<vec3> shapeA, List<vec3> shapeB, CollisionInfo out, vec3 axis, MinMax m1, MinMax m2) {
  var minDistance = double.MAX_FINITE; //Initialize the length of the collision vector to a relatively large value
  var lg = shapeA.length;
  for (int i = 0; i < lg; i++ ) {

//    //This will skip edges that lie totally inside the bodies, as they don't matter.
//    //The boundary flag has to be set manually and defaults to true
//    if (e.boundary == false) {
//      continue;
//    }

    // Calculate the perpendicular to this edge and normalize it
    var v0 = shapeA[i];
    var v1 = (i + 1 == lg)? shapeA[0]: shapeA[i+1];
    axis.x = v0.y - v1.y;
    axis.y = v1.x - v0.x;
    axis.z = 0.0;
    axis.normalize();

    var distance = intervalDistanceOnAxis( shapeA, shapeB, axis);

    // If the intervals don't overlap, return, since there is no collision
    if (distance > 0.0) {
      return false;
    } else if (distance.abs() < minDistance ) {
      minDistance = distance.abs();

      // Save collision information for later
      out.normal.x = axis.x;
      out.normal.y = axis.y;
      //Store the edge, as it is the collision edge
      out.edgev0 = v0;
      out.edgev1 = v1;
    }
  }
  out.depth = minDistance;
  //There is no separating axis. Report a collision!
  return true;
}

void updateCollisionInfo(List<vec3> shapeA, List<vec3> shapeB, CollisionInfo out, vec3 center1, vec3 center2) {
  // int Sign = SGN( CollisionInfo.Normal.multiplyVal( shapeA.Center.minus(shapeB.Center) ) ); //This is needed to make sure that the collision normal is pointing at shapeA
  var xx = center2.x - center1.x;
  var yy = center2.y - center1.y;
  var mult = out.normal.x * xx + out.normal.y * yy;

  // Remember that the line equation is N*( R - R0 ). We choose shapeB->Center as R0; the normal N is given by the collision normal

  if (mult < 0) {
    // Revert the collision normal if it points away from shapeA
    out.normal.x = 0 - out.normal.x;
    out.normal.y = 0 - out.normal.y;
  }

  var smallestD = double.MAX_FINITE; //Initialize the smallest distance to a large value

  for (int i = 0; i < shapeB.length; i++) {
    // Measure the distance of the vertex from the line using the line equation
    // float Distance = CollisionInfo.Normal.multiplyVal( shapeA.Vertices[I].Position.minus(shapeB.Center) );
    xx = shapeB[i].x - center1.x;
    yy = shapeB[i].y - center1.y;
    var distance = out.normal.x * xx + out.normal.y * yy;

    if (distance < smallestD) { //If the measured distance is smaller than the smallest distance reported so far, set the smallest distance and the collision vertex
      smallestD = distance;
      out.v = shapeB[i];
    }
  }
}

vec3 calculateCenter(List<vec3> shape) {

}