library quadtreexy;

import 'package:vector_math/vector_math.dart';

/// see
/// * [Quadtree](http://en.wikipedia.org/wiki/Quadtree) at wikipedia
/// * [JavaScript QuadTree Implementation](http://www.mikechambers.com/blog/2011/03/21/javascript-quadtree-implementation/)
/// * [Quick Tip: Use Quadtrees to Detect Likely Collisions in 2D Space](http://gamedev.tutsplus.com/tutorials/implementation/quick-tip-use-quadtrees-to-detect-likely-collisions-in-2d-space/)
class QuadTreeXY{
  /// (x, y, width, height) bounds of the QuadTree
  final int x, y, w, h;

  /// Whether the QuadTree will contain points (true), or items with bounds (width / height)(false).
  final int maxDepth;
  /// The maximum number of children that a node can contain before it is split into sub-nodes.
  final int maxChildren;

  final _children = new List<vec3>();
  //TODO optimize to reuse nodes (recycle, pool)
  final _nodes = new List<QuadTreeXY>(4);
  var _isLeaf = true;
  var vec2 _splitPoint;

  QuadTreeXY(this.x, this.y, this.w, this.h, [this.maxDepth = 4, this.maxChildren = 4]) {
    _splitPoint = new vec2(w / 2, h / 2);
  }

  /// use [clear] to reset the instance isntead of create a new one with same bounds.
  clear() {
    _children.clear();
    _nodes.forEach((i) => i.clear());
    _isLeaf = true;
  }

  bool insert(vec3 v) {
    var r = _findRegion(v);
    if (r != null) {
      r._insert(v);
    }
    return (r != null);
  }

  QuadTreeXY findRegion(vec3 v) {
    if (v.x < x || v.x > (x + w) || (v.y < y) || v.y > (y + w)) {
      return null;
    }
    return _findRegion(v);
  }

  _insert(vec3 v) {
    _children.add(v);
    if (_children.length > maxChildren && maxDepth > 0) {
      _split();
      _children.forEach((v) => insert(v));
      _children.clear();
    }
  }

  _split() {
    if(_nodes[0] == null) {
      _nodes[0] = new QuadTreeXY(x                , y                , _splitPoint.x    , _splitPoint.y    , maxDepth - 1, maxChildren);
      _nodes[1] = new QuadTreeXY(x + _splitPoint.x, y                , w - _splitPoint.x, _splitPoint.y    , maxDepth - 1, maxChildren);
      _nodes[2] = new QuadTreeXY(x                , y + _splitPoint.y, _splitPoint.x    , h - _splitPoint.y, maxDepth - 1, maxChildren);
      _nodes[3] = new QuadTreeXY(x + _splitPoint.x, y + _splitPoint.y, w - _splitPoint.x, h - _splitPoint.y, maxDepth - 1, maxChildren);
    }
    _isLeaf = false;
  }

  _findRegion(vec3 v) {
    if (_isLeaf) return this;
    var x0 = v.x - x;
    var y0 = v.y - y;
    var b = (x0 <= _splitPoint.x) ?
        ((y0 <= _splitPoint.y) ? _nodes[0] : _nodes[2])
        :((y0 <= _splitPoint.y) ? _nodes[1] : _nodes[3])
        ;
    return b._findChild(v);
  }
}